"""Test VASP parsing utilities."""

from tempfile import TemporaryDirectory
from pathlib import Path

from emmet.core.vasp.utils import (
    recursive_discover_vasp_files,
    discover_and_sort_vasp_files,
    FileMetadata,
)


def test_file_meta(tmp_dir):
    import hashlib
    from monty.io import zopen

    incar_bytes = """
ALGO = Normal
ENCUT = 500
EDIFF = 0.0001
IBRION = -1
""".encode()

    with zopen(file_name := "INCAR.bz2", "wb") as f:
        f.write(incar_bytes)

    file_meta = FileMetadata(name="INCAR.bz2", path=file_name)
    assert Path(file_meta.path).exists()
    assert file_meta.md5 == hashlib.md5(incar_bytes).hexdigest()


def test_file_discovery():
    directory_structure = {
        "./neb_calc/00/": ["INCAR.gz", "KPOINTS", "POSCAR.gz"],
        "./neb_calc/01/": [
            "INCAR.gz",
            "CHGCAR",
            "CONTCAR.gz",
            "KPOINTS",
            "OUTCAR",
            "POSCAR.gz",
            "POTCAR.bz2",
            "vasprun.xml",
        ],
        "./neb_calc/02/": ["INCAR.gz", "KPOINTS", "POSCAR.gz"],
        "neb_calc": ["INCAR.gz", "KPOINTS", "POSCAR.gz", "POTCAR.lzma"],
        "block_2025_02_30/launcher_2025_02_31/launcher_2025_02_31_0001": [
            "AECCAR0.bz2",
            "LOCPOT.gz",
            "CONTCAR.relax1",
            "OUTCAR.relax1",
            "vasprun.xml.relax1",
            "POSCAR.T=300.gz",
            "POSCAR.T=1000.gz",
        ],
    }
    for idx in range(1, 3):
        directory_structure[
            "block_2025_02_30/launcher_2025_02_31/launcher_2025_02_31_0001"
        ].extend(
            [
                f"INCAR.relax{idx}",
                f"INCAR.orig.relax{idx}",
                f"POSCAR.relax{idx}",
                f"POSCAR.orig.relax{idx}",
                f"POTCAR.relax{idx}",
                f"POTCAR.orig.relax{idx}",
            ]
        )
    with TemporaryDirectory() as _tmp_dir:
        tmp_dir = Path(_tmp_dir).resolve()
        for calc_dir, files in directory_structure.items():
            p = tmp_dir / calc_dir
            if not p.exists():
                p.mkdir(parents=True, exist_ok=True)
            for f in files:
                (p / f).touch()
        vasp_files = recursive_discover_vasp_files(tmp_dir)
        valid_vasp_files = recursive_discover_vasp_files(tmp_dir, only_valid=True)

        assert len(recursive_discover_vasp_files(tmp_dir, max_depth=2)) == 4
        assert len(recursive_discover_vasp_files(tmp_dir, max_depth=1)) == 1

        sorted_files = discover_and_sort_vasp_files(
            tmp_dir / "block_2025_02_30/launcher_2025_02_31/launcher_2025_02_31_0001"
        )

    assert set([b.name for b in sorted_files["contcar_file"]]) == {"CONTCAR.relax1"}
    assert set([b.name for b in sorted_files["vasprun_file"]]) == {"vasprun.xml.relax1"}
    assert set([b.name for b in sorted_files["outcar_file"]]) == {"OUTCAR.relax1"}
    assert set([b.name for b in sorted_files["volumetric_files"]]) == {
        "AECCAR0.bz2",
        "LOCPOT.gz",
    }
    assert set([b.name for b in sorted_files["elph_poscars"]]) == {
        "POSCAR.T=300.gz",
        "POSCAR.T=1000.gz",
    }

    assert len(vasp_files) == len(
        directory_structure
    )  # should find all of the defined calculation directories
    for calc_dir, files in directory_structure.items():
        assert (base_path := tmp_dir / calc_dir) in vasp_files
        assert all(f in [b.name for b in vasp_files[base_path]] for f in files)

    # Should only find two valid calculation directories
    valid_calc_dirs = (
        "./neb_calc/01/",
        "block_2025_02_30/launcher_2025_02_31/launcher_2025_02_31_0001",
    )
    assert all(
        f in [b.name for b in valid_vasp_files[tmp_dir / p]]
        for p in valid_calc_dirs
        for f in directory_structure[p]
    )



"""Test phonon document models."""

import numpy as np
from monty.serialization import loadfn
from pymatgen.core import Structure
import pytest

from emmet.core.phonon import PhononDOS, PhononBS, PhononBSDOSDoc
from tests.conftest import assert_schemas_equal

try:
    import pyarrow.parquet as pq
except ImportError:
    pq = None


@pytest.fixture(scope="module")
def legacy_ph_task(test_dir):
    return loadfn(test_dir / "mp-23302_phonon.json.gz", cls=None)


def test_legacy_migration(legacy_ph_task):
    # ensure that legacy phonon data can be migrated to current schema

    assert all(legacy_ph_task.get(k) for k in ("ph_bs", "ph_dos"))
    ph_doc = PhononBSDOSDoc.from_structure(
        Structure.from_dict(legacy_ph_task["ph_bs"]["structure"]), **legacy_ph_task
    )
    assert_schemas_equal(ph_doc, PhononBSDOSDoc.model_config)

    # check remap of phonon DOS
    for k in ("densities", "frequencies"):
        assert np.all(
            np.abs(
                np.array(getattr(ph_doc.phonon_dos, k, []))
                - np.array(legacy_ph_task["ph_dos"].get(k, []))
            )
            < 1e-6
        )

    # check remap of phonon bandstructure
    for k in ("qpoints", "frequencies"):
        assert np.all(
            np.abs(
                np.array(getattr(ph_doc.phonon_bandstructure, k, []))
                - np.array(legacy_ph_task["ph_bs"].get(k, []))
            )
            < 1e-6
        )

    temps = [5, 100, 300, 500, 800]
    ref_data = {
        "entropy": [
            0.10690825310989084,
            64.92923008995054,
            118.49759120562243,
            143.87894281334866,
            167.29013558233882,
        ],
        "heat_capacity": [
            0.3096320518171239,
            47.157121712113714,
            49.56965404227773,
            49.77072796011147,
            49.8399369784691,
        ],
        "internal_energy": [
            1988.2321308731728,
            5269.03083181402,
            15060.035206972329,
            24999.080865250606,
            39943.07283219206,
        ],
        "free_energy": [
            1987.697589607624,
            -1223.8921771810346,
            -20489.242154714408,
            -46940.39054142372,
            -93889.03563367906,
        ],
    }
    tprops = ph_doc.compute_thermo_quantites(temps)
    assert all(
        t == pytest.approx(tprops["temperature"][i]) for i, t in enumerate(temps)
    )
    for k, ref_vals in ref_data.items():
        assert all(
            tprops[k][i] == pytest.approx(ref_val) for i, ref_val in enumerate(ref_vals)
        )

    # check population of structure metadata fields
    assert ph_doc.composition == ph_doc.structure.composition
    assert ph_doc.volume == ph_doc.structure.volume
    assert ph_doc.nelements == len(ph_doc.structure.composition.elements)
    assert sorted(ph_doc.elements) == sorted(ph_doc.structure.composition.elements)


@pytest.mark.skipif(pq is None, reason="pyarrow must be installed to run this test.")
def test_arrow(tmp_dir, legacy_ph_task):
    # test to parquet and rehydration
    ph_doc = PhononBSDOSDoc(**legacy_ph_task)
    arrow_table = ph_doc.objects_to_arrow()
    pq.write_table(arrow_table, "test.parquet")

    rehyd = pq.read_table("test.parquet")

    dos_from_table = PhononDOS.from_arrow(arrow_table, col_prefix="dos_")
    dos_from_parquet = PhononDOS.from_arrow(rehyd, col_prefix="dos_")

    assert ph_doc.phonon_dos == dos_from_table
    assert ph_doc.phonon_dos == dos_from_parquet

    bs_from_table = PhononBS.from_arrow(arrow_table, col_prefix="bs_")
    bs_from_parquet = PhononBS.from_arrow(rehyd, col_prefix="bs_")

    assert ph_doc.phonon_bandstructure == bs_from_table
    assert ph_doc.phonon_bandstructure == bs_from_parquet



"""Define schemas for trajectories."""
from __future__ import annotations

from enum import Enum
import numpy as np
from pydantic import BaseModel, Field
from pathlib import Path
from typing import TYPE_CHECKING

from pymatgen.core import Element, Structure
from pymatgen.core.trajectory import Trajectory as PmgTrajectory

from monty.dev import requires
from monty.serialization import dumpfn

from emmet.core.math import Vector3D, Matrix3D
from emmet.core.tasks import TaskDoc
from emmet.core.vasp.calculation import ElectronicStep

try:
    import pyarrow as pa
    import pyarrow.parquet as pa_pq
    from pyarrow import Table as ArrowTable

except ImportError:
    pa = None
    pa_pq = None
    ArrowTable = None

if TYPE_CHECKING:
    from typing import Any
    from typing_extensions import Self
    from collections.abc import Sequence


class TrajFormat(Enum):
    """Define known trajectory formats."""

    PMG = "json"
    ASE = "traj"
    PARQUET = "parquet"


class Trajectory(BaseModel):
    """Low memory schema for trajectories that can interface with parquet, pymatgen, and ASE."""

    elements: list[int] = Field(
        description="The proton number Z of the elements in the sites"
    )

    cart_coords: list[list[Vector3D]] = Field(
        description="The Cartesian coordinates (Å) of the sites at each ionic step"
    )
    num_ionic_steps: int = Field(description="The number of ionic steps.")

    constant_lattice: Matrix3D | None = Field(
        None,
        description="The constant lattice throughout the trajectory. If not populated, it is assumed that the lattice varies.",
    )
    lattice: list[Matrix3D] | None = Field(
        None,
        description="The lattice at each ionic step. If not populated, it is assumed the lattice is constant through the trajectory.",
    )

    electronic_steps: list[list[ElectronicStep]] | None = Field(
        None, description="The electronic steps within a given ionic step."
    )
    num_electronic_steps: list[int] | None = Field(
        None, description="The number of electronic steps within each ionic step."
    )

    energy: list[float] | None = Field(
        None,
        description="The total energy in eV at each ionic step. This key is also called e_0_energy.",
    )
    e_wo_entrp: list[float] | None = Field(
        None,
        description="The total energy in eV without electronic pseudoentropy from smearing of the Fermi surface.",
    )
    e_fr_energy: list[float] | None = Field(
        None, description="The total electronic free energy in eV."
    )
    forces: list[list[Vector3D]] | None = Field(
        None, description="The interatomic forces in eV/Å at each ionic step."
    )
    stress: list[Matrix3D] | None = Field(
        None, description="The 3x3 stress tensor in kilobar at each ionic step."
    )

    identifier: str | None = Field(
        None, description="Identifier of this trajectory, e.g., task ID."
    )

    def __hash__(self) -> int:
        """Used to verify roundtrip conversion of Trajectory."""
        return hash(self.model_dump_json())

    @staticmethod
    def reorder_structure(structure: Structure, ref_z: list[int]) -> Structure:
        """
        Ensure that the sites in a structure match the order in a set of reference Z values.

        Quick returns if the structure already matches the reference order.

        Parameters
        -----------
        structure : pymatgen .Structure
            Structure to order sites in
        ref_z : list of int
            List of proton numbers / Z values in the reference structure.

        Returns
        -----------
        pymatgen .Structure : structure matching the reference order of sites
        """
        if [site.species.elements[0].Z for site in structure] == ref_z:
            return structure

        running_sites = list(range(len(structure)))
        ordered_sites = []
        for z in ref_z:
            for idx in running_sites:
                if structure[idx].species.elements[0].Z == z:
                    ordered_sites.append(structure[idx])
                    running_sites.remove(idx)
                    break

        if len(running_sites) > 0:
            raise ValueError("Cannot order structure according to reference list.")

        return Structure.from_sites(ordered_sites)

    @classmethod
    def _from_dict(
        cls,
        props: dict[str, Any],
        constant_lattice: bool = False,
        lattice_match_tol: float | None = 1.0e-6,
        **kwargs,
    ) -> Self:
        """
        Common class constructor to create a Trajectory from a dict.

        Parameters
        -----------
        props : dict of str to list
            Dictionary with keys in cls.model_fields and values which are lists
        constant_lattice : bool (default = False)
            Whether the lattice is constant thoughout the trajectory
        lattice_match_tol : float or None (default = 1e-6)
            If a float and constant_lattice is False, this defines the tolerance
            for determining if the lattice has changed during the course of the
            trajectory. If None, no check is performed.
        **kwargs
            Other kwargs to pass to Trajectory

        Returns
        -----------
        Trajectory
        """
        structures = props.pop("structure")
        if any(not structure.is_ordered for structure in structures):
            raise ValueError(
                "At this time, the Trajectory model in emmet does not support disordered structures."
            )

        num_ionic_steps = len(structures)

        props["elements"] = [site.species.elements[0].Z for site in structures[0]]
        if constant_lattice or (
            lattice_match_tol
            and all(
                np.all(
                    np.abs(structures[i].lattice.matrix - structures[j].lattice.matrix)
                    < lattice_match_tol
                )
                for i in range(num_ionic_steps)
                for j in range(i + 1, num_ionic_steps)
            )
        ):
            props["constant_lattice"] = structures[0].lattice.matrix
        else:
            props["lattice"] = [structure.lattice.matrix for structure in structures]

        props["cart_coords"] = [
            cls.reorder_structure(structure, props["elements"]).cart_coords
            for structure in structures
        ]

        if len(esteps := props.get("electronic_steps", [])) > 0:
            props["num_electronic_steps"] = [len(estep) for estep in esteps]

        props["num_ionic_steps"] = num_ionic_steps

        return cls(**props, **kwargs)

    @classmethod
    def from_task_doc(cls, task_doc: TaskDoc, **kwargs) -> Self:
        """
        Create a trajectory from a TaskDoc.

        Parameters
        -----------
        task_doc : emmet.core.TaskDoc
        constant_lattice : bool (default = False)
            Whether the lattice is constant thoughout the trajectory
        kwargs
            Other kwargs to pass to Trajectory

        Returns
        -----------
        Trajectory
        """
        props: dict[str, list] = {
            "structure": [],
            "cart_coords": [],
            "electronic_steps": [],
            "energy": [],
            "e_wo_entrp": [],
            "e_fr_energy": [],
            "forces": [],
            "stress": [],
        }
        # un-reverse the calcs_reversed
        for cr in task_doc.calcs_reversed[::-1]:
            for ionic_step in cr.output.ionic_steps:
                props["structure"].append(ionic_step.structure)

                props["energy"].append(ionic_step.e_0_energy)
                for k in (
                    "e_fr_energy",
                    "e_wo_entrp",
                    "forces",
                    "stress",
                    "electronic_steps",
                ):
                    props[k].append(getattr(ionic_step, k))

        return cls._from_dict(
            props,
            identifier=task_doc.task_id.string if task_doc.task_id else None,
            **kwargs,
        )

    @classmethod
    def from_pmg(cls, traj: PmgTrajectory, **kwargs) -> Self:
        """
        Create a trajectory from a pymatgen .Trajectory.

        Parameters
        -----------
        traj : pymatgen.core.trajectory.Trajectory
        constant_lattice : bool (default = False)
            Whether the lattice is constant thoughout the trajectory
        lattice_match_tol : float or None (default = 1e-6)
            If a float and constant_lattice is False, this defines the tolerance
            for determining if the lattice has changed during the course of the
            trajectory. If None, no check is performed.

        Returns
        -----------
        Trajectory
        """
        props: dict[str, list] = {
            "structure": [structure for structure in traj],
        }
        if traj.frame_properties:
            for k in cls.model_fields:
                vals = [fp.get(k) for fp in traj.frame_properties]
                if all(v is not None for v in vals):
                    props[k] = vals

        return cls._from_dict(props, **kwargs)

    @requires(
        pa is not None, message="pyarrow must be installed to de-/serialize to parquet"
    )
    def to_arrow(
        self, file_name: str | Path | None = None, **write_file_kwargs
    ) -> ArrowTable:
        """
        Create a PyArrow Table from a Trajectory.

        Parameters
        -----------
        file_name : str, .Path, or None (default)
            If not None, a file to write the parquet-format output to.
            Accepts any compression extension used by pyarrow.write_table
        **write_file_kwargs
            If file_name is not None, any kwargs to pass to
            pyarrow.parquet.write_file

        Returns
        -----------
        pyarrow.Table
        """
        pa_config = {
            k: pa.array(v)
            for k, v in self.model_dump(mode="json").items()
            if hasattr(v, "__len__") and not isinstance(v, str)
        }
        pa_config["elements"] = pa.array(
            [self.elements for _ in range(self.num_ionic_steps)]
        )
        pa_config["identifier"] = pa.array(
            [self.identifier for _ in range(self.num_ionic_steps)]
        )

        pa_table = pa.table(pa_config)
        if file_name:
            pa_pq.write_table(pa_table, file_name, **write_file_kwargs)

        return pa_table

    @classmethod
    def from_parquet(cls, file_name: str | Path, identifier: str | None = None) -> Self:
        """
        Create a trajectory from a parquet file.

        Parameters
        -----------
        file_name : str or .Path
            The parquet file to read from
        identifier : str or None (default)
            If not None, the identifier of the trajectory to return
            from a dataset of multiple trajectories.

        Returns
        -----------
        Trajectory
        """
        pa_table = pa_pq.read_table(file_name)
        if identifier:
            pa_table = pa_table.filter(
                pa.compute.field("identifier") == identifier,
                null_selection_behavior="drop",
            )
        config = pa_table.to_pydict()
        config["num_ionic_steps"] = len(config["elements"])
        for k in ("elements", "identifier"):
            config[k] = config[k][0]
        return cls(**config)

    def to_pmg(self, frame_props: Sequence[str] | None = None) -> PmgTrajectory:
        """
        Create a pymatgen Trajectory.

        Parameters
        -----------
        frame_props : Sequence of str or None (default)
            If not None, a list of model fields to populate the frame properties
            of the pymatgen Trajectory with.
            If None, defaults to all available fields.

        Returns
        -----------
        pymatgen.core.trajectory.Trajectory
        """
        frame_props = frame_props or [
            "energy",
            "e_wo_entrp",
            "e_fr_energy",
            "forces",
            "stress",
            "electronic_steps",
        ]

        species = [Element.from_Z(z) for z in self.elements]
        structures = []
        frame_properties = []

        for i, coords in enumerate(self.cart_coords):
            structure = Structure(
                lattice=self.constant_lattice
                if self.constant_lattice
                else self.lattice[i],  # type: ignore[index]
                species=species,
                coords=coords,
                coords_are_cartesian=True,
            )
            structures.append(structure)

            props = {}
            for k in frame_props:
                if (_prop := getattr(self, k, None)) is not None:
                    prop = _prop[i]
                    for cmth in ("model_dump", "tolist"):
                        if hasattr(prop, cmth):
                            prop = getattr(prop, cmth)()
                    props[k] = prop

            frame_properties.append(props)

        return PmgTrajectory.from_structures(
            structures,
            constant_lattice=self.constant_lattice is not None,
            frame_properties=frame_properties,
        )

    def to(
        self,
        file_name: str | Path | None = None,
        fmt: TrajFormat | str | None = None,
        **kwargs,
    ):
        """
        Generic interface to multiple trajectory formats.

        file_name : str, .Path, or None (default)
            If not None, the name of the file to write to.
            If fmt is None, the file format will be inferred from this.
        fmt : TrajFormat, str, or None (default)
            The format of the output trajectory. If file_name is None,
            and fmt is None, defaults to a parquet-compatible
            PyArrow table.
        **kwargs
            Any kwargs supported by Trajectory.to_* methods.
        """
        if file_name and not fmt:
            for _fmt in TrajFormat:
                if _fmt.value in str(file_name).lower():
                    fmt = _fmt
                    break
        elif not file_name and not fmt:
            fmt = TrajFormat.PARQUET

        if isinstance(fmt, str) and fmt.upper() in TrajFormat.__members__:
            fmt = TrajFormat[fmt.upper()]
        else:
            fmt = TrajFormat(fmt)

        if fmt in (TrajFormat.PMG, TrajFormat.ASE):
            traj = self.to_pmg(**kwargs)

            if fmt == TrajFormat.PMG:
                if file_name:
                    dumpfn(traj, file_name)
            elif fmt == TrajFormat.ASE:
                if hasattr(PmgTrajectory, "to_ase"):
                    traj = traj.to_ase(ase_traj_file=file_name)  # type: ignore[attr-defined]
                else:
                    raise ImportError(
                        "Ensure you have pymatgen>=2025.1.23 to use ASE trajectory interfaces."
                    )

        elif fmt == TrajFormat.PARQUET:
            traj = self.to_arrow(file_name=file_name)

        return traj



import numpy as np
from pytest import approx, fixture

from pymatgen.core.trajectory import Trajectory as PmgTraj

from emmet.core.tasks import TaskDoc
from emmet.core.trajectory import Trajectory


@fixture(scope="module")
def si_task(test_dir):
    return TaskDoc.from_directory(test_dir / "vasp" / "Si_old_double_relax")


def test_task_doc(si_task):
    traj = Trajectory.from_task_doc(si_task)
    assert traj.num_ionic_steps == sum(
        len(cr.output.ionic_steps) for cr in si_task.calcs_reversed
    )
    istep = -1
    for cr in si_task.calcs_reversed:
        for ionic_step in cr.output.ionic_steps:
            assert traj.energy[istep] == approx(ionic_step.e_0_energy)
            istep -= 1


def test_parquet(si_task, tmp_dir):
    parqet_file = "test.parquet"

    traj = Trajectory.from_task_doc(si_task)
    traj.to(parqet_file, compression="GZIP")

    new_traj = Trajectory.from_parquet(parqet_file)
    assert hash(new_traj) == hash(traj)


def test_pmg(si_task):
    traj = Trajectory.from_task_doc(si_task)
    pmg_traj = traj.to(fmt="PMG")
    assert isinstance(pmg_traj, PmgTraj)
    assert len(pmg_traj) == traj.num_ionic_steps
    assert len(pmg_traj.frame_properties) == traj.num_ionic_steps

    # some float-rounding that happens here on round trip
    roundtrip = Trajectory.from_pmg(pmg_traj)

    for k in Trajectory.model_fields:
        if (
            k
            in (
                "num_ionic_steps",
                "electronic_steps",
            )
            or getattr(traj, k, None) is None
        ):
            # Can't easily test for equality of electronic steps without bespoke code
            continue

        assert all(
            np.all(np.abs(np.array(new_val) - np.array(getattr(traj, k)[i])) < 1e-6)
            for i, new_val in enumerate(getattr(roundtrip, k))
        )



""" Core definition of a Task Document which represents a calculation from some program"""
from datetime import datetime
from typing import Union, Optional

from pydantic import Field

from emmet.core.base import EmmetBaseModel
from emmet.core.mpid import MPID, MPculeID
from emmet.core.utils import utcnow


class BaseTaskDocument(EmmetBaseModel):
    """
    Definition of base Task Document
    """

    calc_code: str = Field(description="The calculation code used to compute this task")
    version: Optional[str] = Field(
        None, description="The version of the calculation code"
    )
    dir_name: Optional[str] = Field(None, description="The directory for this task")
    task_id: Optional[Union[MPID, MPculeID]] = Field(
        None, description="the Task ID For this document"
    )

    completed: bool = Field(False, description="Whether this calcuation completed")
    completed_at: Optional[datetime] = Field(
        None, description="Timestamp for when this task was completed"
    )
    last_updated: datetime = Field(
        default_factory=utcnow,
        description="Timestamp for when this task document was last updated",
    )



from typing import Dict, Optional
from hashlib import blake2b

from pydantic import Field

from emmet.core.math import Vector3D
from emmet.core.mpid import MPculeID
from emmet.core.material import PropertyOrigin
from emmet.core.qchem.task import TaskDocument
from emmet.core.molecules.molecule_property import PropertyDoc


__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


class ElectricMultipoleDoc(PropertyDoc):
    """Electric multipole (dipole, etc.) properties of a molecule"""

    property_name: str = "multipole_moments"

    total_dipole: float = Field(
        ...,
        description="Total molecular dipole moment (Debye)",
    )

    dipole_moment: Vector3D = Field(
        ...,
        description="Molecular dipole moment vector (Debye)",
    )

    resp_total_dipole: Optional[float] = Field(
        None,
        description="Total dipole moment, calculated via restrained electrostatic potential (RESP) (Debye)",
    )

    resp_dipole_moment: Optional[Vector3D] = Field(
        None,
        description="Molecular dipole moment vector, calculated via RESP (Debye)",
    )

    quadrupole_moment: Optional[Dict[str, float]] = Field(
        None,
        description="Quadrupole moment components (Debye Ang)",
    )

    octopole_moment: Optional[Dict[str, float]] = Field(
        None,
        description="Octopole moment components (Debye Ang^2)",
    )

    hexadecapole_moment: Optional[Dict[str, float]] = Field(
        None,
        description="Hexadecapole moment tensor components (Debye Ang^2)",
    )

    @classmethod
    def from_task(
        cls,
        task: TaskDocument,
        molecule_id: MPculeID,
        deprecated: bool = False,
        **kwargs,
    ):  # type: ignore[override]
        """
        Construct an electric multipole doc
        """

        if task.output.optimized_molecule is not None:
            mol = task.output.optimized_molecule
        else:
            mol = task.output.initial_molecule

        task_type = task.task_type.value

        # For SP and force calcs, dipoles stored in output
        dipoles = task.output.dipoles
        if dipoles is None:
            if task_type in ["Single Point", "Force"]:
                raise Exception("No dipoles in task!")
            total_dipole = None
            dipole_moment = None
            resp_total_dipole = None
            resp_dipole_moment = None
        else:
            total_dipole = dipoles.get("total")
            dipole_moment = dipoles.get("dipole")
            resp_total_dipole = dipoles.get("RESP_total")
            resp_dipole_moment = dipoles.get("RESP_dipole")

        # Look for multipole moment in calcs_reversed
        calcs_reversed = task.calcs_reversed

        # FFOpt/FFTSOpt: grab final multipole moments from final optimization
        if task_type in [
            "Frequency Flattening Geometry Optimization",
            "Frequency Flattening Transition State Geometry Optimization",
        ]:
            calc = calcs_reversed[1]
            grab_index = -1
        # Frequency calc: grab first multipole moments
        # This is necessary if finite difference is used, in which case multipole moments will be generated for each
        # perturbation
        elif task_type == "Frequency Analysis":
            calc = calcs_reversed[0]
            grab_index = 0
        # Otherwise, grab final available multipole moments
        else:
            calc = calcs_reversed[0]
            grab_index = -1

        dipoles = calc.get("dipoles", dict())

        if total_dipole is None:
            total_dipole = dipoles.get("total")
            if isinstance(total_dipole, list):
                total_dipole = total_dipole[grab_index]
        if dipole_moment is None:
            dipole_moment = dipoles.get("dipole")
            if isinstance(dipole_moment, list) and len(dipole_moment) == 0:
                dipole_moment = None
            elif isinstance(dipole_moment[0], list):
                dipole_moment = dipole_moment[grab_index]
        if resp_total_dipole is None:
            resp_total_dipole = dipoles.get("RESP_total")
            if isinstance(resp_total_dipole, list):
                resp_total_dipole = resp_total_dipole[grab_index]
        if resp_dipole_moment is None:
            resp_dipole_moment = dipoles.get("RESP_dipole")
            if isinstance(resp_dipole_moment, list) and len(resp_dipole_moment) == 0:
                resp_dipole_moment = None
            elif isinstance(resp_dipole_moment[0], list):
                resp_dipole_moment = resp_dipole_moment[grab_index]

        if total_dipole is None or dipole_moment is None:
            raise Exception("Total dipole or dipole moment vector missing!")

        multipoles = calc.get("multipoles", dict())
        quadrupoles = multipoles.get("quadrupole")
        octopoles = multipoles.get("octopole")
        hexadecapoles = multipoles.get("hexadecapole")

        quadrupole_moment = None
        octopole_moment = None
        hexadecapole_moment = None

        if quadrupoles is not None:
            if isinstance(quadrupoles, dict):
                quadrupole_moment = quadrupoles
            else:
                quadrupole_moment = quadrupoles[grab_index]

        if octopoles is not None:
            if isinstance(octopoles, dict):
                octopole_moment = octopoles
            else:
                octopole_moment = octopoles[grab_index]

        if hexadecapoles is not None:
            if isinstance(hexadecapoles, dict):
                hexadecapole_moment = hexadecapoles
            else:
                hexadecapole_moment = hexadecapoles[grab_index]

        id_string = f"electricmultipole-{molecule_id}-{task.task_id}-{task.lot_solvent}"
        h = blake2b()
        h.update(id_string.encode("utf-8"))
        property_id = h.hexdigest()

        return super().from_molecule(
            meta_molecule=mol,
            property_id=property_id,
            molecule_id=molecule_id,
            level_of_theory=task.level_of_theory,
            solvent=task.solvent,
            lot_solvent=task.lot_solvent,
            molecule=mol,
            total_dipole=total_dipole,
            dipole_moment=dipole_moment,
            resp_total_dipole=resp_total_dipole,
            resp_dipole_moment=resp_dipole_moment,
            quadrupole_moment=quadrupole_moment,
            octopole_moment=octopole_moment,
            hexadecapole_moment=hexadecapole_moment,
            warnings=list(),  # No warnings currently available
            origins=[PropertyOrigin(name="multipole_moments", task_id=task.task_id)],
            deprecated=deprecated,
            **kwargs,
        )



"""Define schemas for DFPT, phonopy, and pheasy-derived phonon data."""

from __future__ import annotations

from datetime import datetime
from enum import Enum
from functools import cached_property
import json
from pathlib import Path
import yaml  # type: ignore[import-untyped]

from monty.io import zopen
from monty.os.path import zpath
from monty.dev import requires
import numpy as np
from pydantic import model_validator, BaseModel, Field, computed_field, PrivateAttr
from typing import Optional, TYPE_CHECKING

from pymatgen.core import Lattice, Structure
from pymatgen.electronic_structure.bandstructure import Kpoint
from pymatgen.phonon.bandstructure import PhononBandStructureSymmLine
from pymatgen.phonon.dos import PhononDos as PhononDosObject

from emmet.core.base import CalcMeta
from emmet.core.utils import DocEnum
from emmet.core.math import Vector3D, Matrix3D, Tensor4R
from emmet.core.mpid import MPID
from emmet.core.polar import DielectricDoc, BornEffectiveCharges, IRDielectric
from emmet.core.structure import StructureMetadata
from emmet.core.common import convert_datetime
from emmet.core.utils import utcnow, get_num_formula_units

from typing_extensions import Literal

try:
    import pyarrow as pa

    from pyarrow import Table as ArrowTable
except ImportError:
    pa = None
    ArrowTable = None

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any
    from typing_extensions import Self

DEFAULT_PHONON_FILES = {
    "structure": "POSCAR",
    "phonon_bandstructure": "band.yaml",  # chosen as Phonopy default
    "phonon_dos": "total_dos.dat",  # chosen as Phonopy default
    "force_constants": "FORCE_CONSTANTS",  # chosen as Phonopy default
    "born": "born.npz",
    "epsilon_static": "epsilon_static.npz",
    "phonopy_output": "phonopy.yaml",
}


class PhononMethod(Enum):
    """Define common methods for computed phonon properties."""

    DFPT = "dfpt"
    PHONOPY = "phonopy"
    PHEASY = "pheasy"


class PhononDOS(BaseModel):
    """Define schema of pymatgen phonon density of states."""

    frequencies: list[float] = Field(description="The phonon frequencies in THz.")
    densities: list[float] = Field(description="The phonon density of states.")

    @cached_property
    def to_pmg(self) -> PhononDosObject:
        """Get / cache corresponding pymatgen object."""
        return PhononDosObject(frequencies=self.frequencies, densities=self.densities)

    @requires(pa is not None, "`pip install pyarrow` to use this functionality.")
    def to_arrow(self, col_prefix: str | None = None) -> ArrowTable:
        """Convert PhononDOS to a pyarrow Table."""
        col_prefix = col_prefix or ""
        return pa.Table.from_pydict(
            {
                f"{col_prefix}{k}": [getattr(self, k)]
                for k in ("frequencies", "densities")
            }
        )

    @classmethod
    @requires(pa is not None, "`pip install pyarrow` to use this functionality.")
    def from_arrow(cls, table: ArrowTable, col_prefix: str | None = None) -> Self:
        """Create a PhononDOS from a pyarrow Table."""
        col_prefix = col_prefix or ""
        return cls(
            **{k: table[f"{col_prefix}{k}"].to_pylist()[0] for k in cls.model_fields}
        )

    @classmethod
    def from_phonopy(cls, phonon_dos_file: str | Path) -> Self:
        """Create a PhononDOS from phonopy .dat output."""
        phonopy_dos: dict[str, list[float]] = {
            k: []
            for k in (
                "frequencies",
                "densities",
            )
        }
        with zopen(phonon_dos_file, "rt") as f:
            for line in f.read().splitlines():
                non_comment_text = line.split("#")[0]  # type: ignore[arg-type]
                if len(cols := non_comment_text.split()) == 2:
                    phonopy_dos["frequencies"].append(float(cols[0]))
                    phonopy_dos["densities"].append(float(cols[1]))

        return cls(**phonopy_dos)


class PhononBS(BaseModel):
    """Define schema of pymatgen phonon band structure."""

    qpoints: list[Vector3D] = Field(
        description="The q-kpoints at which the band structure was sampled, in direct coordinates.",
    )
    frequencies: list[list[float]] = Field(
        description="The phonon frequencies in THz, with the first index representing the band, and the second the q-point.",
    )
    reciprocal_lattice: Matrix3D = Field(description="The reciprocal lattice.")
    has_nac: bool = Field(
        False,
        description="Whether the calculation includes non-analytical corrections at Gamma.",
    )
    eigendisplacements: list[list[list[tuple[complex, complex, complex]]]] | None = (
        Field(None, description="Phonon eigendisplacements in Cartesian coordinates.")
    )
    labels_dict: dict[str, Vector3D] | None = Field(
        None, description="The high-symmetry labels of specific q-points."
    )
    structure: Structure | None = Field(
        None, description="The structure associated with the calculation."
    )
    _primitive_structure: Structure | None = PrivateAttr(None)

    @model_validator(mode="before")
    def rehydrate(cls, config: Any) -> Any:
        """Ensure fields are correctly populated."""
        if isinstance(egd := config.get("eigendisplacements"), dict) and all(
            egd.get(k) is not None for k in ("real", "imag")
        ):
            config["eigendisplacements"] = (
                np.array(egd["real"]) + 1j * np.array(egd["imag"])
            ).tolist()

        if (struct := config.get("structure")) and not isinstance(struct, Structure):
            config["structure"] = Structure.from_dict(struct)

        # remap legacy fields
        for k, v in {
            "lattice_rec": "reciprocal_lattice",
            "bands": "frequencies",
        }.items():
            if config.get(k):
                config[v] = config.pop(k)

        if isinstance(config["reciprocal_lattice"], dict):
            config["reciprocal_lattice"] = config["reciprocal_lattice"].get("matrix")

        return config

    @property
    def primitive_structure(self) -> Structure | None:
        """Cache primitive structure for use in computing entropy, heat capacity, etc."""
        if self.structure and not self._primitive_structure:
            self._primitive_structure = self.structure.get_primitive_structure()
        return self._primitive_structure

    @cached_property
    def to_pmg(self) -> PhononBandStructureSymmLine:
        """Get / cache corresponding pymatgen object."""
        rlatt = Lattice(self.reciprocal_lattice)
        return PhononBandStructureSymmLine(
            [Kpoint(q, lattice=rlatt).frac_coords for q in self.qpoints],  # type: ignore[misc]
            np.array(self.frequencies),
            rlatt,
            has_nac=self.has_nac,
            eigendisplacements=np.array(self.eigendisplacements),
            structure=self.structure,
            labels_dict={
                k: Kpoint(v, lattice=rlatt).frac_coords
                for k, v in (self.labels_dict or {}).items()
            },
            coords_are_cartesian=False,
        )

    @requires(pa is not None, "`pip install pyarrow` to use this functionality.")
    def to_arrow(self, col_prefix: str | None = None) -> ArrowTable:
        """Convert a PhononBS to an arrow table."""
        config = self.model_dump()
        if structure := config.pop("structure", None):
            config["structure"] = json.dumps(structure.as_dict())

        for k in ("qpoints", "frequencies", "reciprocal_lattice", "eigendisplacements"):
            if (vals := config.pop(k, None)) and k == "eigendisplacements":
                cvals = np.array(vals)
                config["eigendisplacements_real"] = cvals.real.tolist()
                config["eigendisplacements_imag"] = cvals.imag.tolist()
            elif vals:
                rvals = np.array(vals)
                config[k] = rvals.tolist()

        if qpt_labels := config.pop("labels_dict"):
            config["qpoint_labels"] = list(qpt_labels)
            config["qpoint_labelled_points"] = [
                qpt_labels[k] for k in config["qpoint_labels"]
            ]

        col_prefix = col_prefix or ""
        return pa.Table.from_pydict(
            {f"{col_prefix}{k}": [v] for k, v in config.items()}
        )

    @classmethod
    @requires(pa is not None, "`pip install pyarrow` to use this functionality.")
    def from_arrow(cls, table: ArrowTable, col_prefix: str | None = None) -> Self:
        """Create a PhononBS from an arrow table."""
        col_prefix = col_prefix or ""
        config: dict[str, Any] = {}
        for k in (
            "structure",
            "has_nac",
            "qpoints",
            "frequencies",
            "reciprocal_lattice",
            "eigendisplacements_real",
            "qpoint_labels",
        ):
            _k = f"{col_prefix}{k}"
            if _k not in table.column_names:
                continue
            v = table[_k].to_pylist()[0]
            if k == "structure":
                config[k] = Structure.from_dict(json.loads(v))
            elif k in ("qpoints", "frequencies", "reciprocal_lattice"):
                config[k] = np.array(v)
            elif k == "eigendisplacements_real":
                config["eigendisplacements"] = (
                    table[f"{col_prefix}eigendisplacements_real"].to_numpy()[0]
                    + 1.0j * table[f"{col_prefix}eigendisplacements_imag"].to_numpy()[0]
                )
            elif k == "qpoint_labels":
                config["labels_dict"] = dict(
                    zip(v, table[f"{col_prefix}qpoint_labelled_points"].to_pylist()[0])
                )
            else:
                config[k] = v
        return cls(**config)

    @classmethod
    def from_phonopy(cls, phonon_bandstructure_file: str | Path):
        """Create a PhononBS from phonopy .yaml output."""
        with zopen(phonon_bandstructure_file, "rt") as f:
            phonopy_bandstructure = yaml.safe_load(f.read())

        phonopy_bandstructure.update(
            qpoints=[entry["q-position"] for entry in phonopy_bandstructure["phonon"]],
            frequencies=[
                [branch["frequency"] for branch in entry["band"]]
                for entry in phonopy_bandstructure["phonon"]
            ],
        )
        return cls(**phonopy_bandstructure)


class SumRuleChecks(BaseModel):
    """Container class for defining sum rule checks."""

    asr: float | None = Field(
        None, description="The violation of the acoustic sum rule."
    )
    cnsr: float | None = Field(
        None, description="The violation of the charge neutral sum rule."
    )


class PhononBSDOSTask(StructureMetadata):
    """Phonon band structures and density of states data."""

    identifier: str | None = Field(
        None, description="The identifier of this phonon analysis task."
    )

    phonon_method: PhononMethod | None = Field(
        None, description="The method used to calculate phonon properties."
    )

    phonon_bandstructure: PhononBS | None = Field(
        None,
        description="Phonon band structure object.",
    )

    phonon_dos: PhononDOS | None = Field(
        None,
        description="Phonon density of states object.",
    )

    epsilon_static: Matrix3D | None = Field(
        None, description="The high-frequency dielectric constant."
    )

    epsilon_electronic: Matrix3D | None = Field(
        None,
        description="The electronic contribution to the high-frequency dielectric constant.",
    )

    born: list[Matrix3D] | None = Field(
        None,
        description="Born charges, only for symmetrically inequivalent atoms",
    )

    # needed, e.g. to compute Grueneisen parameter etc
    force_constants: list[list[Matrix3D]] | None = Field(
        None, description="Force constants between every pair of atoms in the structure"
    )

    last_updated: datetime = Field(
        default_factory=utcnow,
        description="Timestamp for the most recent calculation for this Material document.",
    )

    sum_rules_breaking: SumRuleChecks | None = Field(
        None,
        description="Deviations from sum rules.",
    )

    structure: Structure | None = Field(
        None, description="Structure used in the calculation."
    )

    total_dft_energy: float | None = Field(
        None, description="total DFT energy in eV/atom."
    )

    volume_per_formula_unit: float | None = Field(
        None, description="volume per formula unit in Angstrom**3."
    )

    formula_units: int | None = Field(None, description="Formula units per cell.")

    supercell_matrix: Matrix3D | None = Field(
        None, description="matrix describing the supercell."
    )
    primitive_matrix: Matrix3D | None = Field(
        None, description="matrix describing relationship to primitive cell."
    )

    code: str | None = Field(
        None, description="String describing the code for the computation."
    )

    post_process_settings: PhononComputationalSettings | None = Field(
        None,
        description="Field including settings for the post processing code, e.g., phonopy.",
    )

    thermal_displacement_data: ThermalDisplacementData | None = Field(
        None,
        description="Includes all data of the computation of the thermal displacements",
    )

    calc_meta: list[CalcMeta] | None = Field(
        None,
        description="Metadata for individual calculations used to build this document.",
    )

    @model_validator(mode="before")
    @classmethod
    def migrate_fields(cls, config: Any) -> Any:
        """Migrate legacy input fields."""

        # This block is for older DFPT data
        for k, v in {
            "ph_dos": "phonon_dos",
            "ph_bs": "phonon_bandstructure",
            "e_total": "epsilon_static",
            "e_electronic": "epsilon_electronic",
            "becs": "born",
        }.items():
            if config.get(k):
                config[v] = config.pop(k)

        # Make sure that the datetime field is properly formatted
        if config.get("last_updated"):
            config["last_updated"] = convert_datetime(cls, config["last_updated"])

        if (ph_bs := config.get("phonon_bandstructure")) and not config.get(
            "structure"
        ):
            if isinstance(ph_bs, PhononBandStructureSymmLine | PhononBS):
                config["structure"] = ph_bs.structure
            else:
                config["structure"] = ph_bs.get("structure")

        # This block is for the migration from atomate2 --> emmet-core schemas
        if config.get("structure"):
            if isinstance(config["structure"], dict):
                config["structure"] = Structure.from_dict(config["structure"])

            old_formula_units = config.pop("formula_units", 1)
            if toten := config.pop("total_dft_energy", None):
                config["total_energy"] = (
                    toten * old_formula_units / config["structure"].num_sites
                )

            config["formula_units"] = get_num_formula_units(
                config["structure"].composition
            )

        if (fc := config.get("force_constants")) and isinstance(fc, dict):
            config["force_constants"] = fc["force_constants"]

        if comp_sett := config.get("phonopy_settings"):
            config["post_process_settings"] = comp_sett

        calc_meta_migrate = {
            "uuids": "_uuid",
            "jobdirs": "_job_dir",
        }
        calc_meta_remap = {"uuids": "identifier", "jobdirs": "dir_name"}
        if any(config.get(k) for k in calc_meta_migrate):
            calc_meta: dict[str, dict[str, str]] = {}
            for field_name, splitter in calc_meta_migrate.items():
                if vals := config.get(field_name):
                    if not isinstance(vals, dict):
                        vals = vals.model_dump()
                    for k, v in vals.items():
                        job_name = k.split(splitter)[0]
                        if isinstance(v, list):
                            for idx, sub_v in enumerate(v):
                                if (sub_name := f"{job_name}-{idx}") not in calc_meta:
                                    calc_meta[sub_name] = {}
                                calc_meta[sub_name][calc_meta_remap[field_name]] = sub_v
                        else:
                            if job_name not in calc_meta:
                                calc_meta[job_name] = {}
                            calc_meta[job_name][calc_meta_remap[field_name]] = v

            config["calc_meta"] = [
                CalcMeta(
                    name=job_name,
                    **meta,
                )
                for job_name, meta in calc_meta.items()
            ]

        return config

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def has_imaginary_modes(self) -> bool | None:
        tol: float = 1e-5
        if self.phonon_bandstructure:
            return self.phonon_bandstructure.to_pmg.has_imaginary_freq(tol=tol)
        elif self.phonon_dos:
            return bool(
                np.any(
                    np.array(self.phonon_dos.densities)[
                        np.array(self.phonon_dos.frequencies) < tol
                    ]
                    > tol
                )
            )
        return None

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def charge_neutral_sum_rule(self) -> Matrix3D | None:
        """Sum of Born effective charges over sites should be zero."""
        if self.born:
            bec = np.array(self.born)
            return tuple(tuple(row) for row in np.sum(bec, axis=0).tolist())
        return None

    @computed_field  # type: ignore[prop-decorator]
    @cached_property
    def acoustic_sum_rule(self) -> Matrix3D | None:
        """Sum of q=0 atomic force constants should be zero."""
        if self.force_constants:
            return tuple(
                tuple(row)
                for row in np.einsum("iijk->jk", np.array(self.force_constants))
            )
        return None

    @computed_field  # type: ignore[prop-decorator]
    @property
    def check_sum_rule_deviations(self) -> SumRuleChecks:
        """Report deviations from sum rules."""
        if not self.sum_rules_breaking:
            self.sum_rules_breaking = SumRuleChecks(
                **{
                    k: (
                        np.max(np.abs(getattr(self, attr)))
                        if getattr(self, attr)
                        else None
                    )
                    for k, attr in {
                        "asr": "acoustic_sum_rule",
                        "cnsr": "charge_neutral_sum_rule",
                    }.items()
                }
            )
        return self.sum_rules_breaking

    def _get_thermo_from_dos(
        self,
        quantity: Literal["entropy", "cv", "internal_energy", "helmholtz_free_energy"],
        temperature: float,
        normalization: Literal["atoms", "formula_units"] | None = "formula_units",
    ) -> float | None:
        """Get a thermodynamic property from the phonon DOS.

        quantity : "entropy", "cv", "internal_energy", or "helmholtz_free_energy"
            If "entropy" or "cv", the units before normalization are J/(K * mol).
            If "internal_energy" or "helmholtz_free_energy", the units before normalization are J/mol.
        temperature : float
        normalization : "atoms", "formula_units", or None
            Whether to normalize by the number of atoms in the cell ("atoms"),
            the number of formula units ("formula_units", default), or not (None).
        """

        if not self.phonon_dos:
            return None

        if normalization and not self.structure:
            raise ValueError(
                "Cannot normalize thermodynamic quantities without a structure."
            )

        norm_fac = 1.0
        if normalization == "atoms":
            norm_fac = 1.0 / self.structure.num_sites  # type: ignore[union-attr]
        elif normalization == "formula_units":
            norm_fac = 1.0 / self.formula_units  # type: ignore[operator]
        else:
            raise ValueError(f"Unknown {normalization=} convention.")

        return (
            getattr(self.phonon_dos.to_pmg, quantity)(
                temp=temperature,
            )
            * norm_fac
        )

    def entropy(
        self,
        temperature: float,
        normalization: Literal["atoms", "formula_units"] | None = "formula_units",
    ) -> float | None:
        """Compute the entropy in J/(K * mol * formula units).

        temperature : float
        normalization : "atoms", "formula_units", or None
            Whether to normalize by the number of atoms in the cell ("atoms"),
            the number of formula units ("formula_units"), or not (None).
        """
        return self._get_thermo_from_dos(
            "entropy", temperature, normalization=normalization
        )

    def heat_capacity(
        self,
        temperature: float,
        normalization: Literal["atoms", "formula_units"] | None = "formula_units",
    ) -> float | None:
        """Compute the heat capacity in J/(K * mol * formula units).

        temperature : float
        normalization : "atoms", "formula_units", or None
            Whether to normalize by the number of atoms in the cell ("atoms"),
            the number of formula units ("formula_units"), or not (None).
        """
        return self._get_thermo_from_dos("cv", temperature, normalization=normalization)

    def internal_energy(
        self,
        temperature: float,
        normalization: Literal["atoms", "formula_units"] | None = "formula_units",
    ) -> float | None:
        """Compute the internal energy in J/(mol * formula units).

        temperature : float
        normalization : "atoms", "formula_units", or None
            Whether to normalize by the number of atoms in the cell ("atoms"),
            the number of formula units ("formula_units"), or not (None).
        """
        return self._get_thermo_from_dos(
            "internal_energy", temperature, normalization=normalization
        )

    def free_energy(
        self,
        temperature: float,
        normalization: Literal["atoms", "formula_units"] | None = "formula_units",
    ) -> float | None:
        """Compute the Helmholtz free energy in J/(mol * formula units).

        temperature : float
        normalization : "atoms", "formula_units", or None
            Whether to normalize by the number of atoms in the cell ("atoms"),
            the number of formula units ("formula_units"), or not (None).
        """
        return self._get_thermo_from_dos(
            "helmholtz_free_energy", temperature, normalization=normalization
        )

    def compute_thermo_quantities(
        self,
        temperatures: Sequence[float],
        normalization: Literal["atoms", "formula_units"] | None = "formula_units",
    ) -> dict[str, Sequence[float | None]]:
        """Compute all thermodynamic quantities as a convenience method."""

        quantities = {
            "entropy": "entropy",
            "heat_capacity": "cv",
            "internal_energy": "internal_energy",
            "free_energy": "helmholtz_free_energy",
        }

        thermo_props: dict[str, Sequence[float | None]] = {
            k: [
                self._get_thermo_from_dos(v, temp, normalization=normalization)  # type: ignore[arg-type]
                for temp in temperatures
            ]
            for k, v in quantities.items()
        }
        thermo_props["temperature"] = temperatures
        return thermo_props

    @requires(pa is not None, "`pip install pyarrow` to use this functionality.")
    def objects_to_arrow(self) -> ArrowTable:
        """Convert band structure and DOS to pyarrow table row."""
        table = pa.Table.from_pydict({"material_id": [self.material_id]})
        if self.phonon_bandstructure:
            bst = self.phonon_bandstructure.to_arrow(col_prefix="bs_")
            for k in bst.column_names:
                table = table.append_column(k, bst[k])

        if self.phonon_dos:
            dost = self.phonon_dos.to_arrow(col_prefix="dos_")

            for k in dost.column_names:
                table = table.append_column(k, dost[k])
        return table

    @classmethod
    def from_phonopy_pheasy_files(
        cls,
        structure_file: str | Path,
        phonon_bandstructure_file: str | Path | None = None,
        phonon_dos_file: str | Path | None = None,
        force_constants_file: str | Path | None = None,
        born_file: str | Path | None = None,
        epsilon_static_file: str | Path | None = None,
        phonopy_output_file: str | Path | None = None,
        **kwargs,
    ) -> Self:
        """
        Create a PhononBSDOSDoc from a list of explicit Phonopy/Pheasy file paths.
        """

        cls_config: dict[str, Any] = {
            "structure": Structure.from_file(structure_file),
        }
        if "poscar" in str(structure_file).lower():
            cls_config["code"] = "vasp"

        if phonon_bandstructure_file:
            cls_config["phonon_bandstructure"] = PhononBS.from_phonopy(
                phonon_bandstructure_file
            )

        if phonon_dos_file:
            cls_config["phonon_dos"] = PhononDOS.from_phonopy(phonon_dos_file)

        if force_constants_file:
            # read FORCE_CONSTANTS manually
            force_constant_matrix: np.ndarray
            idxs: tuple[int | None, int | None] = (
                None,
                None,
            )
            irow = 0
            with zopen(force_constants_file, "rt") as f:
                for idx, line in enumerate(f.read().splitlines()):
                    vals = line.strip().split()
                    if idx == 0:
                        force_constant_matrix = np.zeros(
                            (int(vals[0]), int(vals[1]), 3, 3)
                        )
                    elif len(vals) == 2:
                        # idxs written like this for mypy
                        idxs = (
                            int(vals[0]) - 1,
                            int(vals[1]) - 1,
                        )
                        irow = 0
                    elif len(vals) == 3:
                        force_constant_matrix[idxs[0], idxs[1], irow] = [
                            float(v) for v in vals
                        ]
                        irow += 1
            cls_config["force_constants"] = force_constant_matrix.tolist()

        if born_file:
            cls_config["born"] = np.load(born_file)
        if epsilon_static_file:
            cls_config["epsilon_static"] = np.load(epsilon_static_file)

        if phonopy_output_file:
            with zopen(phonopy_output_file, "rt") as f:
                phonopy_output = yaml.safe_load(f.read())
            for k in ("primitive_matrix", "supercell_matrix"):
                cls_config[k] = phonopy_output.get(k)

        return cls.from_structure(cls_config["structure"], **cls_config, **kwargs)

    @classmethod
    def from_phonopy_pheasy_directory(
        cls,
        phonon_dir: Path | str,
        **kwargs,
    ) -> Self:
        """Create a PhononBSDOSDoc from a Phonopy/Pheasy directory.

        Parameters
        -----------
        phonon_dir : str or Path
        **kwargs to pass to `PhononBSDOSDoc.from_phonopy_pheasy_files`
        """
        phonon_path = Path(phonon_dir).resolve()
        file_paths = {}
        for k, file_name in DEFAULT_PHONON_FILES.items():
            if (file_path := Path(zpath(str(phonon_path / file_name)))).exists():
                file_paths[f"{k}_file"] = file_path
        return cls.from_phonopy_pheasy_files(**file_paths, **kwargs)


class PhononBSDOSDoc(PhononBSDOSTask):
    """Built data version of PhononBSDOSTask."""

    material_id: MPID | None = Field(
        None,
        description="The Materials Project ID of the material, of the form mp-******.",
    )
    task_ids: list[str] | None = Field(
        None, description="A list of identifiers that were used to build this document."
    )

    @model_validator(mode="after")
    def match_id_fields(self) -> Self:
        """Ensure that `material_id` aliases inherited `identifier` field."""
        self.identifier = self.material_id
        return self


class PhononComputationalSettings(BaseModel):
    """Collection to store computational settings for the phonon computation."""

    # could be optional and implemented at a later stage?
    npoints_band: int | None = Field(
        None, description="number of points for band structure computation"
    )
    kpath_scheme: str | None = Field(None, description="indicates the kpath scheme")
    kpoint_density_dos: int | None = Field(
        None,
        description="number of points for computation of free energies and densities of states",
    )


class ThermalDisplacementData(BaseModel):
    """Collection to store information on the thermal displacement matrices."""

    freq_min_thermal_displacements: float | None = Field(
        None,
        description="cutoff frequency in THz to avoid numerical issues in the "
        "computation of the thermal displacement parameters",
    )
    thermal_displacement_matrix_cif: list[list[Matrix3D]] | None = Field(
        None, description="field including thermal displacement matrices in CIF format"
    )
    thermal_displacement_matrix: list[list[Matrix3D]] | None = Field(
        None,
        description="field including thermal displacement matrices in Cartesian "
        "coordinate system",
    )
    temperatures_thermal_displacements: list[float] | None = Field(
        None,
        description="temperatures at which the thermal displacement matrices"
        "have been computed",
    )


class PhononWarnings(DocEnum):
    ASR = "ASR break", "acoustic sum rule max breaking is larger than 30 cm^-1."
    CNSR = "CNSR break", "charge neutrality sum rule max breaking is larger than 0.2."
    NEG_FREQ = (
        "has negative frequencies",
        "phonon band structure has negative "
        "frequencies anywhere in the Brillouin zone.",
    )
    SMALL_Q_NEG_FREQ = (
        "has small q negative frequencies",
        "the phonon band structure has negative frequencies,"
        " but these are small and very close to the Gamma point "
        "(usually related to numerical errors).",
    )


class PhononWebsiteBS(BaseModel):
    """
    Document with a serialized version of the phonon band structure suitable
    for the phononwebsite (http://henriquemiranda.github.io/phononwebsite/).
    """

    material_id: str = Field(
        ...,
        description="The ID of this material, used as a universal reference across property documents."
        "This comes in the form: mp-******",
    )

    doc_type: Literal["phononwebsite"] = Field(
        "phononwebsite",
        description="The type of the document: a phonon band structure for the phononwebsite.",
    )

    phononwebsite: dict | None = Field(
        None,
        description="Phononwebsite dictionary to plot the animated " "phonon modes.",
    )

    last_updated: datetime = Field(
        description="Timestamp for the most recent calculation update for this property",
        default_factory=datetime.utcnow,
    )

    created_at: datetime = Field(
        description="Timestamp for when this material document was first created",
        default_factory=datetime.utcnow,
    )


class Ddb(BaseModel):
    """
    Document with a the string version of the DDB file produced by abinit.
    """

    material_id: str = Field(
        ...,
        description="The ID of this material, used as a universal reference across property documents."
        "This comes in the form: mp-******",
    )

    doc_type: Literal["ddb"] = Field(
        "ddb", description="The type of the document: a DDB file."
    )

    ddb: str | None = Field(None, description="The string of the DDB file.")

    last_updated: datetime = Field(
        description="Timestamp for the most recent calculation update for this property",
        default_factory=datetime.utcnow,
    )

    created_at: datetime = Field(
        description="Timestamp for when this material document was first created",
        default_factory=datetime.utcnow,
    )


class ThermodynamicProperties(BaseModel):
    """
    Definition of the thermodynamic properties extracted from the phonon frequencies.
    """

    temperatures: list[float] = Field(
        ...,
        description="The list of temperatures at which the thermodynamic properties "
        "are calculated",
    )

    cv: list[float] = Field(
        ...,
        description="The values of the constant-volume specific heat.",
        alias="heat_capacity",
    )

    entropy: list[float] = Field(
        ..., description="The values of the vibrational entropy."
    )


class VibrationalEnergy(BaseModel):
    """
    Definition of the vibrational contribution to the energy as function of
    the temperature.
    """

    temperatures: list[float] = Field(
        ...,
        description="The list of temperatures at which the thermodynamic properties "
        "are calculated",
    )

    internal_energy: list[float] = Field(
        ..., description="The values of the phonon contribution to the internal energy."
    )

    helmholtz_free_energy: list[float] = Field(
        ..., description="The values of the Helmholtz free energy."
    )

    zero_point_energy: float = Field(
        ..., description="The value of the zero point energy."
    )


class Phonon(StructureMetadata):
    """
    Definition for a document with data produced by a phonon calculation.
    """

    material_id: str = Field(
        ...,
        description="The ID of this material, used as a universal reference across property documents."
        "This comes in the form: mp-******",
    )

    structure: Structure = Field(
        ..., description="The relaxed structure for the phonon calculation."
    )

    asr_break: Optional[float] = Field(
        None, description="The maximum breaking of the acoustic sum rule (ASR)."
    )

    warnings: Optional[list[PhononWarnings]] = Field(
        None, description="List of warnings associated to the phonon calculation."
    )

    dielectric: Optional[DielectricDoc] = Field(
        None, description="Dielectric properties obtained during a phonon calculations."
    )

    becs: Optional[BornEffectiveCharges] = Field(
        None, description="Born effective charges obtained for a phonon calculation."
    )

    ir_spectra: Optional[IRDielectric] = Field(
        None, description="The IRDielectricTensor."
    )

    thermodynamic: Optional[ThermodynamicProperties] = Field(
        None,
        description="The thermodynamic properties extracted from the phonon "
        "frequencies.",
    )

    vibrational_energy: Optional[VibrationalEnergy] = Field(
        None, description="The vibrational contributions to the total energy."
    )

    last_updated: datetime = Field(
        description="Timestamp for when this document was last updated",
        default_factory=datetime.utcnow,
    )

    created_at: datetime = Field(
        description="Timestamp for when this material document was first created",
        default_factory=datetime.utcnow,
    )


class AbinitPhonon(Phonon):
    """
    Definition for a document with data produced from a phonon calculation
    with Abinit.
    """

    abinit_input_vars: Optional[dict] = Field(
        None,
        description="Dict representation of the inputs used to obtain the phonon"
        "properties and the main general options (e.g. number of "
        "k-points, number of q-points).",
    )


class SoundVelocity(BaseModel):
    """
    Definition for a document with the sound velocities of the acoustic modes
    close to Gamma, as obtained from a phonon calculation.
    """

    material_id: str = Field(
        ...,
        description="The ID of this material, in the form: mp-******",
    )

    structure: Structure = Field(
        ..., description="The relaxed structure for the phonon calculation."
    )

    directions: list[Vector3D] = Field(
        ...,
        description="Q-points identifying the directions for the calculation"
        "of the speed of sound. In fractional coordinates.",
    )

    labels: list[Optional[str]] = Field(..., description="labels of the directions.")

    sound_velocities: list[Vector3D] = Field(
        ...,
        description="Values of the sound velocities in SI units.",
    )

    mode_types: list[tuple[Optional[str], Optional[str], Optional[str]]] = Field(
        ...,
        description="The types of the modes ('transversal', 'longitudinal'). "
        "None if not correctly identified.",
    )

    last_updated: datetime = Field(
        description="Timestamp for when this document was last updated",
        default_factory=datetime.utcnow,
    )

    created_at: datetime = Field(
        description="Timestamp for when this material document was first created",
        default_factory=datetime.utcnow,
    )


class ThermalDisplacement(BaseModel):
    """
    Definition of a Document for the generalized density of states and
    mean square displacements related to phonon oscillations.
    """

    material_id: str = Field(
        ...,
        description="The ID of this material, used as a universal reference across property documents."
        "This comes in the form: mp-******",
    )

    last_updated: datetime = Field(
        description="Timestamp for the most recent calculation update for this property",
        default_factory=datetime.utcnow,
    )

    created_at: datetime = Field(
        description="Timestamp for when this material document was first created",
        default_factory=datetime.utcnow,
    )

    nsites: int = Field(
        ...,
        description="The number of sites in the structure.",
    )

    nomega: int = Field(
        ...,
        description="The number of frequencies.",
    )

    ntemp: int = Field(
        ...,
        description="The number of temperatures for which the displacements are calculated",
    )

    temperatures: list[float] = Field(
        ...,
        description="The list of temperatures at which the thermodynamic properties "
        "are calculated",
    )

    frequencies: list[float] = Field(
        ..., description="The list of frequencies for the generalized DOS"
    )

    gdos_aijw: Tensor4R = Field(
        ...,
        description=" Generalized DOS in Cartesian coords, with shape (nsites, 3, 3, nomega)",
    )

    amu: dict = Field(
        ..., description="Dictionary of the atomic masses in atomic units."
    )

    structure: Structure = Field(
        ..., description="The relaxed structure for the phonon calculation."
    )

    ucif_t: Tensor4R = Field(
        ...,
        description="Mean squared displacement U tensors as a function of T for T in tmesh in CIF format."
        "With shape (natom, 3, 3, ntemp) ",
    )
    ucif_string_t300k: str = Field(
        ...,
        description="Mean squared displacement U tensors at T=300K in CIF string format.",
    )
